# MyPhone.py - A simple way to access your iPhone music from your desktop
# By Adam Geitgey <ageitgey@gmail.com>
#
# This program is released under the GNU GPL, version 2 or later.
# See http://www.gnu.org/licenses/gpl.html
# 
# This script incorporates code from:
#
# - latin1_to_ascii -- The UNICODE Hammer -- AKA "The Stupid American"
#   by Noah Spurrier
#   http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/251871
#
# - moveMP3 version 2.2 (Uses a modified ReadSongInfo() function)
#   copyright April 2006 by Mike Bleyer <mike @ hoc dot net>
#   http://www.hoc.net/~mike/source/python/
#
# This is meant to make your life easier 
# (or at least the 10 or so people who possess both an iPhone and the 
#  knowledge to install this script and it's prereqs)
#
# Have fun and don't steal music! Etc!



import BaseHTTPServer
import SimpleHTTPServer
import urllib
import os
import sys,os,string,struct

verbose = 10

iTunesDBPath = "/var/root/Media/iTunes_Control/iTunes/iTunesDB"
#iTunesDBPath = "iTunesDB"

class ReqHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):

	def do_HEAD(self):
		self.send_response(200)
		self.send_header("Content-type", "text/html")
		self.end_headers()
		
		
	def do_GET(self):
		#Get the raw request string
		request = self.path[1:]
		
		requestDict = {}
		
		#Look for params
		x = request.find("?")
		if x > -1:
			params = request[x+1:]
			
			paramList = params.split("&")
			
			for param in paramList:
				pair = param.split("=")
				if len(pair) > 1:
					requestDict[urllib.unquote_plus(pair[0])] = urllib.unquote_plus(pair[1])

			if requestDict.has_key("get"):
				filename = "/var/root/Media" + requestDict["get"].replace(":", "/").replace("\"", "")
				print "Getting file: " + filename
				
				ctype = self.guess_type(filename)
				if ctype.startswith('text/'):
					mode = 'r'
				else:
					mode = 'rb'
				try:
					f = open(filename, mode)
					self.send_response(200)
					self.send_header("Content-type", ctype)
					fs = os.fstat(f.fileno())
					self.send_header("Content-Length", str(fs[6]))
					self.send_header("Last-Modified", self.date_time_string(fs.st_mtime))
					self.end_headers()
					self.copyfile(f, self.wfile)
					f.close()			
				except IOError:
					self.send_error(404, "File not found")
		else:
			self.send_response(200)
			self.wfile.write('Content-type: text/html\n\n')
			#if no params, serve up base UI page		
			self.wfile.write(self.getBasePage())

	# Dear non-English-speaking-world,
	#   I'm sorry in advance.  Yes, this is lame.  I know that.
	# - The U.S.A
	#
	# Legal note:
	# The latin1_to_ascii function was written by Noah Spurrier
	def latin1_to_ascii (self, unicrap):
		"""This takes a UNICODE string and replaces Latin-1 characters with
			something equivalent in 7-bit ASCII. It returns a plain ASCII string. 
			This function makes a best effort to convert Latin-1 characters into 
			ASCII equivalents. It does not just strip out the Latin-1 characters.
			All characters in the standard 7-bit ASCII range are preserved. 
			In the 8th bit range all the Latin-1 accented letters are converted 
			to unaccented equivalents. Most symbol characters are converted to 
			something meaningful. Anything not converted is deleted.
		"""
		xlate={0xc0:'A', 0xc1:'A', 0xc2:'A', 0xc3:'A', 0xc4:'A', 0xc5:'A',
			0xc6:'Ae', 0xc7:'C',
			0xc8:'E', 0xc9:'E', 0xca:'E', 0xcb:'E',
			0xcc:'I', 0xcd:'I', 0xce:'I', 0xcf:'I',
			0xd0:'Th', 0xd1:'N',
			0xd2:'O', 0xd3:'O', 0xd4:'O', 0xd5:'O', 0xd6:'O', 0xd8:'O',
			0xd9:'U', 0xda:'U', 0xdb:'U', 0xdc:'U',
			0xdd:'Y', 0xde:'th', 0xdf:'ss',
			0xe0:'a', 0xe1:'a', 0xe2:'a', 0xe3:'a', 0xe4:'a', 0xe5:'a',
			0xe6:'ae', 0xe7:'c',
			0xe8:'e', 0xe9:'e', 0xea:'e', 0xeb:'e',
			0xec:'i', 0xed:'i', 0xee:'i', 0xef:'i',
			0xf0:'th', 0xf1:'n',
			0xf2:'o', 0xf3:'o', 0xf4:'o', 0xf5:'o', 0xf6:'o', 0xf8:'o',
			0xf9:'u', 0xfa:'u', 0xfb:'u', 0xfc:'u',
			0xfd:'y', 0xfe:'th', 0xff:'y',
			0xa1:'!', 0xa2:'{cent}', 0xa3:'{pound}', 0xa4:'{currency}',
			0xa5:'{yen}', 0xa6:'|', 0xa7:'{section}', 0xa8:'{umlaut}',
			0xa9:'{C}', 0xaa:'{^a}', 0xab:'<<', 0xac:'{not}',
			0xad:'-', 0xae:'{R}', 0xaf:'_', 0xb0:'{degrees}',
			0xb1:'{+/-}', 0xb2:'{^2}', 0xb3:'{^3}', 0xb4:"'",
			0xb5:'{micro}', 0xb6:'{paragraph}', 0xb7:'*', 0xb8:'{cedilla}',
			0xb9:'{^1}', 0xba:'{^o}', 0xbb:'>>', 
			0xbc:'{1/4}', 0xbd:'{1/2}', 0xbe:'{3/4}', 0xbf:'?',
			0xd7:'*', 0xf7:'/'
			}
	
		r = ''
		for i in unicrap:
			if xlate.has_key(ord(i)):
				r += xlate[ord(i)]
			elif ord(i) >= 0x80:
				pass
			else:
				r += str(i)
		return r


	def getBasePage(self):
		global iTunesDBPath
		html = ""

		html += "<html><head><title>MyPhone</title></head><body>"
		html += "MyPhone: Your iPhone's Music Collection<br><i>Music is shared for your personal use only.</i><br><table>"

		songs = self.ReadSongInfo(iTunesDBPath)
		
		#sort it to make it pretty
		keys = songs.keys()
		keys.sort()
		
		lastGroupTitle = ""
		
		for file in keys:
			artist = songs[file][0]
			album = songs[file][1]
			track = songs[file][2]
			trackNum = songs[file][4]
			fileName = songs[file][5]
			
			groupTitle = artist + " - " + album
			if groupTitle != lastGroupTitle:
				lastGroupTitle = groupTitle
				html += "</table><br><table><td colspan=2><b>" + groupTitle + "</b></td>"

			html += "<tr>"
			
			if trackNum != "" and trackNum != "0":
				html += "<td align=right width=20>" + trackNum + ". </td>"
			else:
				html += "<td width=20>&nbsp;</td>"
			html += "<td><a href=?get=\""+fileName+"\">"
			html += track
			html += "</a></td></tr>"

		html += "</body></html>"
		
		return html
		
	# Legal Note:
	# The ReadSongInfo function is written by Mike Bleyer and modified by Adam Geitgey
	def ReadSongInfo(self, iTunesDBFilename='iTunesDB'):
		"""Read a dictionary of items from the db file. Return format is 'filename':('Artist','Album','Song')"""
		# init empty dict
		songdict={}
		# make sure db file exists
		if not os.path.isfile(iTunesDBFilename):
			sys.stderr.write('Error: iTunesDB file \"'+iTunesDBFilename+'\" not found.\n')
			return songdict
		if verbose>0:
			sys.stdout.write('Reading iTunesDB file \"'+iTunesDBFilename+'\".\n')
		# read the song data
		bigstring=''
		try:
			dbFile=open(iTunesDBFilename,'rb')
			bigstring=dbFile.read()
			dbFile.close()
		except:
			sys.stderr.write('Error: cannot open iTunesDB file \"'+iTunesDBFilename+'\" for reading.\n')
			return songdict
	
		# parse the iTunes db file info, it's in UTF-16 Unicode (little endian)
		# this is still somewhat clumsy, it works for now
		# and has the advantage that we don't care about the exact iTunesDB record format
		if verbose>0:
			sys.stdout.write('Parsing iTunesDB file \"'+iTunesDBFilename+'\" for MP3 tracks.\n')
		fileHeader='mhit'
		itemHeader='mhod'
		# the record structure is as follows:
		# a file record (mhit) contains info about a single song file
		# it has several string item child records (mhod) which contain info such as:
		#    song title, artist, album, genre, filetype, filename
		max=len(bigstring)-5
		i=0
		while i<max:
			# we found a filename item
			if bigstring[i:i+4]==fileHeader:
				# seed with defaults, not every song must have all these set
				# but we require them
				mp3Filename=None
				mp3Artist=u'Artist'
				mp3Album=u'Album'
				mp3Title=u'Title'
				mp3Genre=u'Genre'
				# we have a song file item, get its header length
				(songHeadLen,)=struct.unpack('<L',bigstring[i+4:i+8])
				fieldformat = '<LLLLLBBBBLLLLLLLLLLLLLLLLLLLL8sBBHHHLLL'
				#             ' 44444111144444444444444444444 811222444422...'
				field_bytes=struct.calcsize(fieldformat)
	
				vals = (struct.unpack(fieldformat, bigstring[i+8:i+8+field_bytes]))
				songCompLen = vals[0]
				songItems = vals[1]
				songKey = vals[2]
				visible = vals[3]
				songFileType = vals[4]
				ft1 = vals[5]
				ft2 = vals[6]
				songCompilation_flag = vals[7]
				songStars_rating = vals[8]
				songfile_date = vals[9]
				songfile_size = vals[10]
				songfile_duration = vals[11]
				songPosition = vals[12]
				songNumber = vals[13]
				songYear = vals[14]
				songBitrate = vals[15]
				songu12 = vals[16]
				songsample_rate1 = vals[17]
				songvolume_adjust = vals[18]
				songstartplayback = vals[19]
				songstopplayback = vals[20]
				songsoundcheck = vals[21]
				songplay_count1 = vals[22]
				songplay_count2 = vals[23]
				songlast_time_played = vals[24]
				songdisc_number = vals[25]
				songtotal_discs = vals[26]
				songapplestore_userid = vals[27]
				songlast_modified = vals[28]
				songbookmark_time = vals[29]
				songdb_song_id = vals[30]
				songchecked = vals[31]
				songapplication_rating = vals[32]
				songbpm = vals[33]
				songartwork_count = vals[34]
				u9 = vals[35]
				songartwork_size = vals[36]
				u11 = vals[37]
	
				# length of the complete song file record
				#(songCompLen,)=struct.unpack('<L',bigstring[i+8:i+12])
				# number of string items
				#(songItems,)=struct.unpack('<L',bigstring[i+12:i+16])
				# song position on album
				#(songPosition,)=struct.unpack('<L',bigstring[i+44:i+48])
				# number of songs on album
				#(songNumber,)=struct.unpack('<L',bigstring[i+48:i+52])
				# song year
				#(songYear,)=struct.unpack('<L',bigstring[i+52:i+56])
				# song bitrate
				#(songBitrate,)=struct.unpack('<L',bigstring[i+56:i+60])
				# skip to start of first item
				i=i+songHeadLen
				# now read all the string item records
				j=0
				for j in range(songItems):
					# until the next filename item or end of list
					if (bigstring[i:i+4]==itemHeader):
						# we have a string item, get its full length
						(itemCompLen,)=struct.unpack('<L',bigstring[i+8:i+12])
						# get its type
						(itemType,)=struct.unpack('<L',bigstring[i+12:i+16])
						# get string length
						(itemLen,)=struct.unpack('<L',bigstring[i+28:i+32])
						# set i to the beginning of next item record
						i=i+itemCompLen
						# get the bytes, if it's really a string depends on type
						itemArray=bigstring[i-itemLen:i]
						# convert to string depending on type
						if itemType==1:
							mp3Title=unicode(itemArray,'utf-16-le')
						elif itemType==2:
							mp3Filename=unicode(itemArray,'utf-16-le')
						elif itemType==3:
							mp3Album=unicode(itemArray,'utf-16-le')
						elif itemType==4:
							mp3Artist=unicode(itemArray,'utf-16-le')
						elif itemType==5:
							mp3Genre=unicode(itemArray,'utf-16-le')
					else:
						sys.stderr.write('Error parsing iTunesDB string records (mhod)\n')
				# safety check number of mhod records
				if j!=(songItems-1):
					sys.stderr.write('Error parsing iTunesDB: number of mhod strings not correct\n')
				# add song file info to dict, if we have a valid file pathname
				if mp3Filename:
					#songdict[mp3Filename]=(mp3Artist,mp3Album,mp3Title,mp3Genre)
					songdict[self.latin1_to_ascii(mp3Artist + "-" + mp3Album+"-%03d-%03d)" % (songNumber, songPosition) + mp3Title)]=(self.latin1_to_ascii(mp3Artist),self.latin1_to_ascii(mp3Album),self.latin1_to_ascii(mp3Title),self.latin1_to_ascii(mp3Genre), str(songPosition), mp3Filename)
				# set counter back by one, so we don't skip the next song file
				i=i-1
			i=i+1
		return songdict
		
	
httpd = BaseHTTPServer.HTTPServer(("", 1111), ReqHandler)
httpd.serve_forever()	